@@ -49,7 +49,6 @@
 #include <paths.h>
 #include <pwd.h>
 #include <regex.h>
-#include <search.h>
 #include <stdarg.h>
 #include <stdint.h>
 #include <stdio.h>
@@ -73,6 +72,15 @@
 #define FNM_CASEFOLD FNM_IGNORECASE
 #endif
 
+/* For Hurd. */
+#if !defined(PATH_MAX)
+#define PATH_MAX 4096
+#endif
+
+struct fitree;
+struct idtree;
+
+static int Bflag;
 static int Cflag;
 static char *Cflags[64];
 static int Gflag;
@@ -92,7 +100,6 @@ static char *argv0;
 static char *format;
 static char *ordering;
 static struct expr *expr;
-static void *root = NULL; // tree
 static int prune;
 static size_t prefixl;
 static char input_delim = '\n';
@@ -108,15 +115,13 @@ static char long_format[] = "%M%x %n %u %g %s %\324F %\324R %p%F%l\n";
 static char zero_format[] = "%p\\0";
 static char stat_format[] = "%D %i %M %n %u %g %R %s \"%Ab %Ad %AT %AY\" \"%Tb %Td %TT %TY\" \"%Cb %Cd %CT %CY\" %b %p\n";
 
-static void *users;
-static void *groups;
-static void *filesystems;
-static int scanned_filesystems;
+static struct fitree *root;
+static struct fitree *new_root;
 
-struct idmap {
-	long id;
-	char *name;
-};
+static struct idtree *users;
+static struct idtree *groups;
+static struct idtree *filesystems;
+static int scanned_filesystems;
 
 static int need_stat;
 static int need_group;
@@ -134,6 +139,7 @@ static off_t maxsize;
 static blkcnt_t maxblocks;
 static unsigned int maxxattr;
 
+static int bflag_depth;
 static int maxdepth;
 static int uwid, gwid, fwid;
 
@@ -162,6 +168,7 @@ struct fileinfo {
 enum op {
 	EXPR_OR = 1,
 	EXPR_AND,
+	EXPR_COND,
 	EXPR_NOT,
 	EXPR_LT,
 	EXPR_LE,
@@ -227,7 +234,7 @@ struct expr {
 		char *string;
 		int64_t num;
 		regex_t *regex;
-	} a, b;
+	} a, b, c;
 };
 
 static char *pos;
@@ -399,7 +406,7 @@ parse_op()
 }
 
 static struct expr *parse_cmp();
-static struct expr *parse_or();
+static struct expr *parse_cond();
 
 static struct expr *
 parse_inner()
@@ -410,6 +417,11 @@ parse_inner()
 	} else if (token("print")) {
 		struct expr *e = mkexpr(EXPR_PRINT);
 		return e;
+	} else if (token("skip")) {
+		struct expr *e = mkexpr(EXPR_PRINT);
+		struct expr *not = mkexpr(EXPR_NOT);
+		not->a.expr = e;
+		return not;
 	} else if (token("color")) {
 		struct expr *e = mkexpr(EXPR_COLOR);
 		int64_t n;
@@ -426,7 +438,7 @@ parse_inner()
 		not->a.expr = e;
 		return not;
 	} else if (token("(")) {
-		struct expr *e = parse_or();
+		struct expr *e = parse_cond();
 		if (token(")"))
 			return e;
 		parse_error("missing ) at '%.15s'", pos);
@@ -866,11 +878,34 @@ parse_or()
 	return r;
 }
 
+static struct expr *
+parse_cond()
+{
+	struct expr *e1 = parse_or();
+
+	if (token("?")) {
+		struct expr *e2 = parse_or();
+		if (token(":")) {
+			struct expr *e3 = parse_cond();
+			struct expr *r = mkexpr(EXPR_COND);
+			r->a.expr = e1;
+			r->b.expr = e2;
+			r->c.expr = e3;
+
+			return r;
+		} else {
+			parse_error(": expected at '%.15s'", pos);
+		}
+	}
+
+	return e1;
+}
+
 static struct expr *
 parse_expr(const char *s)
 {
 	pos = (char *)s;
-	struct expr *e = parse_or();
+	struct expr *e = parse_cond();
 	if (*pos)
 		parse_error("trailing garbage at '%.15s'", pos);
 	return e;
@@ -897,27 +932,94 @@ static const char *
 readlin(const char *p, const char *alt)
 {
 	static char b[PATH_MAX];
-	int r = readlink(p, b, sizeof b - 1);
-	if (r < 0)
+	ssize_t r = readlink(p, b, sizeof b - 1);
+	if (r < 0 || (size_t)r >= sizeof b - 1)
 		return alt;
 	b[r] = 0;
 	return b;
 }
 
-int
-idorder(const void *a, const void *b)
-{
-	struct idmap *ia = (struct idmap *)a;
-	struct idmap *ib = (struct idmap *)b;
+//// AA-tree implementation, adapted from https://github.com/ccxvii/minilibs
+struct idtree {
+	long id;
+	char *name;
+	struct idtree *left, *right;
+	int level;
+};
 
-	if (ia->id == ib->id)
-		return 0;
-	else if (ia->id < ib->id)
-		return -1;
-	else
-		return 1;
+static struct idtree idtree_sentinel = { 0, 0, &idtree_sentinel, &idtree_sentinel, 0 };
+
+static struct idtree *
+idtree_make(long id, char *name)
+{
+	struct idtree *node = malloc(sizeof (struct idtree));
+	node->id = id;
+	node->name = name;
+	node->left = node->right = &idtree_sentinel;
+	node->level = 1;
+	return node;
 }
 
+char *
+idtree_lookup(struct idtree *node, long id)
+{
+	if (node) {
+		while (node != &idtree_sentinel) {
+			if (id == node->id)
+				return node->name;
+			else if (id < node->id)
+				node = node->left;
+			else
+				node = node->right;
+		}
+	}
+
+	return 0;
+}
+
+static struct idtree *
+idtree_skew(struct idtree *node)
+{
+	if (node->left->level == node->level) {
+		struct idtree *save = node;
+		node = node->left;
+		save->left = node->right;
+		node->right = save;
+	}
+	return node;
+}
+
+static struct idtree *
+idtree_split(struct idtree *node)
+{
+	if (node->right->right->level == node->level) {
+		struct idtree *save = node;
+		node = node->right;
+		save->right = node->left;
+		node->left = save;
+		node->level++;
+	}
+	return node;
+}
+
+struct idtree *
+idtree_insert(struct idtree *node, long id, char *name)
+{
+	if (node && node != &idtree_sentinel) {
+		if (id == node->id)
+			return node;
+		else if (id < node->id)
+			node->left = idtree_insert(node->left, id, name);
+		else
+			node->right = idtree_insert(node->right, id, name);
+		node = idtree_skew(node);
+		node = idtree_split(node);
+		return node;
+	}
+	return idtree_make(id, name);
+}
+////
+
 static char *
 strid(long id)
 {
@@ -929,47 +1031,41 @@ strid(long id)
 static char *
 groupname(gid_t gid)
 {
-	struct idmap key, **result;
-	key.id = gid;
-	key.name = 0;
+	char *name = idtree_lookup(groups, gid);
 
-	if (!(result = tfind(&key, &groups, idorder))) {
-		struct group *g = getgrgid(gid);
-		if (g) {
-			struct idmap *newkey = malloc(sizeof (struct idmap));
-			newkey->id = gid;
-			newkey->name = strdup(g->gr_name);
-			if ((int)strlen(g->gr_name) > gwid)
-				gwid = strlen(g->gr_name);
-			tsearch(newkey, &groups, idorder);
-			return newkey->name;
-		}
+	if (name)
+		return name;
+
+	struct group *g = getgrgid(gid);
+	if (g) {
+		if ((int)strlen(g->gr_name) > gwid)
+			gwid = strlen(g->gr_name);
+		char *name = strdup(g->gr_name);
+		groups = idtree_insert(groups, gid, name);
+		return name;
 	}
 
-	return result ? (*result)->name : strid(gid);
+	return strid(gid);
 }
 
 static char *
 username(uid_t uid)
 {
-	struct idmap key, **result;
-	key.id = uid;
-	key.name = 0;
+	char *name = idtree_lookup(users, uid);
 
-	if (!(result = tfind(&key, &users, idorder))) {
-		struct passwd *p = getpwuid(uid);
-		if (p) {
-			struct idmap *newkey = malloc(sizeof (struct idmap));
-			newkey->id = uid;
-			newkey->name = strdup(p->pw_name);
-			if ((int)strlen(p->pw_name) > uwid)
-				uwid = strlen(p->pw_name);
-			tsearch(newkey, &users, idorder);
-			return newkey->name;
-		}
+	if (name)
+		return name;
+	
+	struct passwd *p = getpwuid(uid);
+	if (p) {
+		if ((int)strlen(p->pw_name) > uwid)
+			uwid = strlen(p->pw_name);
+		char *name = strdup(p->pw_name);
+		users = idtree_insert(users, uid, name);
+		return name;
 	}
-
-	return result ? (*result)->name : strid(uid);
+	
+	return strid(uid);
 }
 
 #if defined(__linux__) || defined(__CYGWIN__)
@@ -985,17 +1081,16 @@ scan_filesystems()
 	 * this will fail if we are not allowed to read the mountpoint, but then
 	 * we should not have to look up this st_dev value... */
 	mtab = setmntent(_PATH_MOUNTED, "r");
+	if (!mtab && errno == ENOENT)
+		mtab = setmntent("/proc/mounts", "r");
 	if (!mtab)
 		return;
 
 	while ((mnt = getmntent(mtab))) {
 		if (stat(mnt->mnt_dir, &st) < 0)
 			continue;
-
-		struct idmap *newkey = malloc(sizeof (struct idmap));
-		newkey->id = st.st_dev;
-		newkey->name = strdup(mnt->mnt_type);
-		tsearch(newkey, &filesystems, idorder);
+		filesystems = idtree_insert(filesystems,
+		    st.st_dev, strdup(mnt->mnt_type));
 	};
 
 	endmntent(mtab);
@@ -1018,11 +1113,8 @@ scan_filesystems()
 	while (getmntent(mtab, &mnt) == 0) {
 		if (stat(mnt.mnt_mountp, &st) < 0)
 			continue;
-
-		struct idmap *newkey = malloc(sizeof (struct idmap));
-		newkey->id = st.st_dev;
-		newkey->name = strdup(mnt.mnt_fstype);
-		tsearch(newkey, &filesystems, idorder);
+		filesystems = idtree_insert(filesystems,
+		    st.st_dev, strdup(mnt->mnt_fstype));
 	};
 
 	fclose(mtab);
@@ -1043,11 +1135,8 @@ scan_filesystems()
 	while (i-- > 0) {
 		if (stat(mnt->f_mntonname, &st) < 0)
 			continue;
-
-		struct idmap *newkey = malloc(sizeof (struct idmap));
-		newkey->id = st.st_dev;
-		newkey->name = strdup(mnt->f_fstypename);
-		tsearch(newkey, &filesystems, idorder);
+		filesystems = idtree_insert(filesystems,
+		    st.st_dev, strdup(mnt->f_fstypename));
 		mnt++;
 	};
 
@@ -1066,11 +1155,8 @@ scan_filesystems()
 	while (i-- > 0) {
 		if (stat(mnt->f_mntonname, &st) < 0)
 			continue;
-
-		struct idmap *newkey = malloc(sizeof (struct idmap));
-		newkey->id = st.st_dev;
-		newkey->name = strdup(mnt->f_fstypename);
-		tsearch(newkey, &filesystems, idorder);
+		filesystems = idtree_insert(filesystems,
+		    st.st_dev, strdup(mnt->f_fstypename));
 		mnt++;
 	};
 
@@ -1087,18 +1173,17 @@ scan_filesystems()
 static char *
 fstype(dev_t devid)
 {
-	struct idmap key, **result;
-	key.id = devid;
-	key.name = 0;
-
 	if (!scanned_filesystems)
 		scan_filesystems();
-	result = tfind(&key, &filesystems, idorder);
 
-	if (result && (int)strlen((*result)->name) > fwid)
-		fwid = strlen((*result)->name);
+	char *name = idtree_lookup(filesystems, devid);
+	if (name) {
+		if ((int)strlen(name) > fwid)
+			fwid = strlen(name);
+		return name;
+	}
 
-	return result ? (*result)->name : strid(devid);
+	return strid(devid);
 }
 
 static char *
@@ -1177,11 +1262,18 @@ count_entries(struct fileinfo *fi)
 int
 eval(struct expr *e, struct fileinfo *fi)
 {
+	long v = 0;
+	const char *s = "";
+
 	switch (e->op) {
 	case EXPR_OR:
 		return eval(e->a.expr, fi) || eval(e->b.expr, fi);
 	case EXPR_AND:
 		return eval(e->a.expr, fi) && eval(e->b.expr, fi);
+	case EXPR_COND:
+		return eval(e->a.expr, fi)
+		    ? eval(e->b.expr, fi)
+		    : eval(e->c.expr, fi);
 	case EXPR_NOT:
 		return !eval(e->a.expr, fi);
 	case EXPR_PRUNE:
@@ -1199,8 +1291,7 @@ eval(struct expr *e, struct fileinfo *fi)
 	case EXPR_GE:
 	case EXPR_GT:
 	case EXPR_ALLSET:
-	case EXPR_ANYSET: {
-		long v = 0;
+	case EXPR_ANYSET:
 		switch (e->a.prop) {
 		case PROP_ATIME: v = fi->sb.st_atime; break;
 		case PROP_CTIME: v = fi->sb.st_ctime; break;
@@ -1216,10 +1307,8 @@ eval(struct expr *e, struct fileinfo *fi)
 		case PROP_SIZE: v = fi->sb.st_size; break;
 		case PROP_TOTAL: v = fi->total; break;
 		case PROP_UID: v = fi->sb.st_uid; break;
-		default:
-			parse_error("unknown property");
+		default: parse_error("unknown property");
 		}
-
 		switch (e->op) {
 		case EXPR_LT: return v < e->b.num;
 		case EXPR_LE: return v <= e->b.num;
@@ -1229,11 +1318,11 @@ eval(struct expr *e, struct fileinfo *fi)
 		case EXPR_GT: return v > e->b.num;
 		case EXPR_ALLSET: return (v & e->b.num) == e->b.num;
 		case EXPR_ANYSET: return (v & e->b.num) > 0;
+		default: parse_error("invalid operator");
 		}
-	}
 	case EXPR_CHMOD:
 		return test_chmod(e->b.string, fi->sb.st_mode & 07777);
-	case EXPR_TYPE: {
+	case EXPR_TYPE:
 		switch (e->a.filetype) {
 		case TYPE_BLOCK: return S_ISBLK(fi->sb.st_mode);
 		case TYPE_CHAR: return S_ISCHR(fi->sb.st_mode);
@@ -1242,15 +1331,14 @@ eval(struct expr *e, struct fileinfo *fi)
 		case TYPE_REGULAR: return S_ISREG(fi->sb.st_mode);
 		case TYPE_SOCKET: return S_ISSOCK(fi->sb.st_mode);
 		case TYPE_SYMLINK: return S_ISLNK(fi->sb.st_mode);
+		default: parse_error("invalid file type");
 		}
-	}
 	case EXPR_STREQ:
 	case EXPR_STREQI:
 	case EXPR_GLOB:
 	case EXPR_GLOBI:
 	case EXPR_REGEX:
-	case EXPR_REGEXI: {
-		const char *s = "";
+	case EXPR_REGEXI:
 		switch (e->a.prop) {
 		case PROP_FSTYPE: s = fstype(fi->sb.st_dev); break;
 		case PROP_GROUP: s = groupname(fi->sb.st_gid); break;
@@ -1259,6 +1347,7 @@ eval(struct expr *e, struct fileinfo *fi)
 		case PROP_TARGET: s = readlin(fi->fpath, ""); break;
 		case PROP_USER: s = username(fi->sb.st_uid); break;
 		case PROP_XATTR: s = xattr_string(fi->fpath); break;
+		default: parse_error("unknown property");
 		}
 		switch (e->op) {
 		case EXPR_STREQ: return strcmp(e->b.string, s) == 0;
@@ -1267,8 +1356,8 @@ eval(struct expr *e, struct fileinfo *fi)
 		case EXPR_GLOBI: return fnmatch(e->b.string, s, FNM_CASEFOLD) == 0;
 		case EXPR_REGEX:
 		case EXPR_REGEXI: return regexec(e->b.regex, s, 0, 0, 0) == 0;
+		default: parse_error("invalid operator");
 		}
-	}
 	default:
 		parse_error("invalid operation %d, please file a bug.", e->op);
 	}
@@ -1378,6 +1467,100 @@ order(const void *a, const void *b)
 	return strcmp(fa->fpath, fb->fpath);
 }
 
+static void
+free_fi(struct fileinfo *fi)
+{
+	if (fi)
+		free(fi->fpath);
+	free(fi);
+}
+
+//// AA-tree implementation, adapted from https://github.com/ccxvii/minilibs
+struct fitree {
+	struct fileinfo *fi;
+	struct fitree *left, *right;
+	int level;
+};
+
+static struct fitree fitree_sentinel = { 0, &fitree_sentinel, &fitree_sentinel, 0 };
+
+static struct fitree *
+fitree_make(struct fileinfo *fi)
+{
+	struct fitree *node = malloc(sizeof (struct fitree));
+	node->fi = fi;
+	node->left = node->right = &fitree_sentinel;
+	node->level = 1;
+	return node;
+}
+
+static struct fitree *
+fitree_skew(struct fitree *node)
+{
+	if (node->left->level == node->level) {
+		struct fitree *save = node;
+		node = node->left;
+		save->left = node->right;
+		node->right = save;
+	}
+	return node;
+}
+
+static struct fitree *
+fitree_split(struct fitree *node)
+{
+	if (node->right->right->level == node->level) {
+		struct fitree *save = node;
+		node = node->right;
+		save->right = node->left;
+		node->left = save;
+		node->level++;
+	}
+	return node;
+}
+
+struct fitree *
+fitree_insert(struct fitree *node, struct fileinfo *fi)
+{
+	if (node && node != &fitree_sentinel) {
+		int c = order(fi, node->fi);
+		if (c == 0)
+			return node;
+		if (c < 0)
+			node->left = fitree_insert(node->left, fi);
+		else
+			node->right = fitree_insert(node->right, fi);
+		node = fitree_skew(node);
+		node = fitree_split(node);
+		return node;
+	}
+	return fitree_make(fi);
+}
+
+void
+fitree_walk(struct fitree *node, void (*visit)(struct fileinfo *))
+{
+	if (!node)
+		return;
+	if (node->left != &fitree_sentinel)
+		fitree_walk(node->left, visit);
+	visit(node->fi);
+	if (node->right != &fitree_sentinel)
+		fitree_walk(node->right, visit);
+}
+
+void
+fitree_free(struct fitree *node)
+{
+	if (node && node != &fitree_sentinel) {
+		fitree_free(node->left);
+		fitree_free(node->right);
+		free_fi(node->fi);
+		free(node);
+	}
+}
+////
+
 static int
 intlen(intmax_t i)
 {
@@ -1821,29 +2004,24 @@ print_format(struct fileinfo *fi)
 	}
 }
 
-void
-print(const void *nodep, const VISIT which, const int depth)
-{
-	(void)depth;
-
-	if (which == postorder || which == leaf)
-		print_format(*(struct fileinfo **)nodep);
-}
+static int initial;
 
 int
-callback(const char *fpath, const struct stat *sb, int depth, int entries, off_t total)
+callback(const char *fpath, const struct stat *sb, int depth, ino_t entries, off_t total)
 {
 	struct fileinfo *fi = malloc(sizeof (struct fileinfo));
 	fi->fpath = strdup(fpath);
 	fi->prefixl = prefixl;
-	fi->depth = depth;
+	fi->depth = Bflag ? (depth > 0 ? bflag_depth + 1 : 0) : depth;
 	fi->entries = entries;
 	fi->total = total;
 	fi->color = current_color;
 	memcpy((char *)&fi->sb, (char *)sb, sizeof (struct stat));
 
-	if (expr && !eval(expr, fi))
+	if (expr && !eval(expr, fi)) {
+		free_fi(fi);
 		return 0;
+	}
 
 	if (need_xattr) {
 		strncpy(fi->xattr, xattr_string(fi->fpath), sizeof fi->xattr);
@@ -1852,11 +2030,23 @@ callback(const char *fpath, const struct stat *sb, int depth, int entries, off_t
 	} else
 		memset(fi->xattr, 0, sizeof fi->xattr);
 
-	if (Uflag)
-		print(&fi, postorder, depth);
-	else
-		// add to the tree, note that this will elimnate duplicate files
-		tsearch(fi, &root, order);
+	if (Uflag) {
+		print_format(fi);
+		free_fi(fi);
+		return 0;
+	} else if (Bflag) {
+		if (initial && fi->depth == 0) {
+			root = fitree_insert(root, fi);
+		} else if (!initial && fi->depth == bflag_depth + 1) {
+			new_root = fitree_insert(new_root, fi);
+		} else {
+			free_fi(fi);
+			return 0;
+		}
+	} else {
+		// add to the tree, note that this will eliminate duplicate files
+		root = fitree_insert(root, fi);
+	}
 
 	if (depth > maxdepth)
 		maxdepth = depth;
@@ -1906,12 +2096,16 @@ recurse(char *path, struct history *h, int guessdir)
 	size_t l = strlen(path), j = l && path[l-1] == '/' ? l - 1 : l;
 	struct stat st;
 	struct history new;
-	int r, entries;
+	int r;
+	ino_t entries;
 	const char *fpath = *path ? path : ".";
 
 	int resolve = Lflag || (Hflag && !h);
 	int root = (path[0] == '/' && path[1] == 0);
 
+	if (Bflag && h && h->chain)
+		return 0;
+
 	if (need_stat)
 		guessdir = 1;
 
@@ -2005,6 +2199,18 @@ recurse(char *path, struct history *h, int guessdir)
 	return 0;
 }
 
+void
+tree_recurse(struct fileinfo *fi)
+{
+	if (S_ISDIR(fi->sb.st_mode)) {
+		char path[PATH_MAX];
+
+		strcpy(path, fi->fpath);
+		bflag_depth = fi->depth;
+		recurse(path, 0, 0);
+	}
+}
+
 int
 traverse_file(FILE *file)
 {
@@ -2055,7 +2261,7 @@ traverse(const char *path)
 	}
 
 	prefixl = strlen(path);
-	while (path[prefixl-1] == '/')
+	while (prefixl && path[prefixl-1] == '/')
 		prefixl--;
 	if (prefixl > PATH_MAX) {
 		errno = ENAMETOOLONG;
@@ -2088,11 +2294,12 @@ main(int argc, char *argv[])
 
 	setlocale(LC_ALL, "");
 
-	while ((c = getopt(argc, argv, "01AC:DFGHLQST:UXde:f:lho:st:x")) != -1)
+	while ((c = getopt(argc, argv, "01ABC:DFGHLQST:UXde:f:lho:st:x")) != -1)
 		switch (c) {
 		case '0': format = zero_format; input_delim = 0; Qflag = 0; break;
-		case '1': expr = chain(parse_expr("depth == 0 || prune"), EXPR_AND, expr); break;
-		case 'A': expr = chain(expr, EXPR_AND, parse_expr("!(path ~~ \"*/.*\" && prune) && path != \".\"")); break;
+		case '1': expr = chain(parse_expr("depth > 0 ? prune : print"), EXPR_AND, expr); break;
+		case 'A': expr = chain(expr, EXPR_AND, parse_expr("name =~ \"^\\.\" && path != \".\" ? prune : print")); break;
+		case 'B': Bflag++; Dflag = 0; Uflag = 0; need_stat++; break;
 		case 'C':
 			if ((unsigned int)Cflag <
 			    sizeof Cflags / sizeof Cflags[0]) {
@@ -2103,7 +2310,7 @@ main(int argc, char *argv[])
 			}
 			Gflag += 2;  // force color on
 			break;
-		case 'D': Dflag++; break;
+		case 'D': Dflag++; Bflag = 0; break;
 		case 'F': format = type_format; break;
 		case 'G': Gflag++; break;
 		case 'H': Hflag++; break;
@@ -2111,7 +2318,7 @@ main(int argc, char *argv[])
 		case 'Q': Qflag++; break;
 		case 'S': Qflag++; format = stat_format; break;
 		case 'T': Tflag = timeflag(optarg); break;
-		case 'U': Uflag++; break;
+		case 'U': Uflag++; Bflag = 0; break;
 		case 'X': Xflag++; break;
 		case 'd': expr = chain(parse_expr("type == d && prune || print"), EXPR_AND, expr); break;
 		case 'e': expr = chain(expr, EXPR_AND,
@@ -2127,7 +2334,7 @@ main(int argc, char *argv[])
 		case 'x': xflag++; break;
 		default:
 			fprintf(stderr,
-"Usage: %s [-0|-F|-l [-TA|-TC|-TM]|-S|-f FMT] [-D] [-H|-L] [-1AGQdhsx]\n"
+"Usage: %s [-0|-F|-l [-TA|-TC|-TM]|-S|-f FMT] [-B|-D] [-H|-L] [-1AGQdhsx]\n"
 "          [-U|-o ORD] [-e REGEX]* [-t TEST]* [-C [COLOR:]PATH]* PATH...\n", argv0);
 			exit(2);
 		}
@@ -2142,6 +2349,16 @@ main(int argc, char *argv[])
 	}
 
 	analyze_format();
+	if (Uflag) {
+		maxnlink = 99;
+		maxsize = 4*1024*1024;
+		maxblocks = maxsize / 512;
+		maxrdev = maxdev = 255;
+		maxuid = maxgid = 65536;
+		maxino = 9999999;
+		maxdepth = 99;
+		uwid = gwid = fwid = 8;
+	}
 	if (Xflag) {
 		basepath = realpath(".", 0);
 		if (!basepath) {
@@ -2175,15 +2392,28 @@ main(int argc, char *argv[])
 
 	current_color = -1;
 
+	initial = 1;
 	if (!Cflag && optind == argc) {
 		traverse("");
 	} else {
 		for (i = optind; i < argc; i++)
 			traverse(argv[i]);
 	}
+	initial = 0;
 
-	if (!Uflag)
-		twalk(root, print);
+	if (Bflag) {
+		while (root) {
+			fitree_walk(root, print_format);
+			fitree_walk(root, tree_recurse);
+			fitree_free(root);
+
+			root = new_root;
+			new_root = 0;
+		}
+	} else if (!Uflag) {
+		fitree_walk(root, print_format);
+		// no need to destroy here, we are done
+	}
 
 	return status;
 }
